/*-
 * Copyright (c) 2008, Alexandre Snarskii
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: aspf.c,v 1.3 2008/10/27 13:35:47 snar Exp $
 */
#include <sys/types.h>
#include <sys/cdefs.h>
#include <sys/systm.h>
#include <sys/mbuf.h>
#include <sys/param.h>
#include <sys/module.h>
#include <sys/kernel.h>
#include <sys/socket.h>
#include <sys/sysctl.h>

#include <net/if.h>
#include <net/pfil.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/udp.h>

#include "aspf.h"

int aspf_modevent(module_t md, int /*modeventtype_t*/ what, void* unused);
int aspf_sysctl_hook(SYSCTL_HANDLER_ARGS);

static unsigned aspf_dns_port = 53;
static unsigned long aspf_cnt_incomplete = 0;
static unsigned long aspf_invalid_dns_q = 0, aspf_valid_dns_q = 0;
static int aspf_enabled=1;

SYSCTL_DECL(_net_inet);
SYSCTL_NODE(_net_inet, OID_AUTO, aspf, CTLFLAG_RW, 0, 
	"Application-Specific packet filters");
SYSCTL_UINT(_net_inet_aspf, OID_AUTO, dns_port, CTLFLAG_RW, &aspf_dns_port, 0, 
	"DNS port");
SYSCTL_ULONG(_net_inet_aspf, OID_AUTO, dnsq_ok, CTLFLAG_RD, 
	&aspf_valid_dns_q, 0, "Valid DNS queries");
SYSCTL_ULONG(_net_inet_aspf, OID_AUTO, dnsq_fail, CTLFLAG_RD, 
	&aspf_invalid_dns_q, 0, "Invalid DNS queries dropped");
SYSCTL_ULONG(_net_inet_aspf, OID_AUTO, incomplete, CTLFLAG_RD,
	&aspf_cnt_incomplete, 0, "Too short packets");
SYSCTL_PROC(_net_inet_aspf, OID_AUTO, enable, CTLTYPE_INT|CTLFLAG_RW,
	&aspf_enabled, 0, aspf_sysctl_hook, "I", "Enable ASPF");

static int 
aspf_check_in(void* unused, struct mbuf **m0, struct ifnet* ifp, int dir,
	struct inpcb* inp)
{ 
	struct ip* h=NULL;
	struct mbuf* m=*m0;
	int off, ulen;
	struct udphdr uh;

	if(!aspf_enabled) 
		return 0;

	if (!m || !(m->m_flags & M_PKTHDR)) 
		return 0;

	if (m->m_pkthdr.len < (int)sizeof(struct ip)) 
		return 0;
	
	h=mtod(m, struct ip*);

	if(h->ip_p != IPPROTO_UDP) 
		return 0;

	if(h->ip_off & htons(IP_MF|IP_OFFMASK)) { 
		/* do not even try to work with fragments. At least for now */
		return 0;
	};

	off=h->ip_hl << 2;

	if(off < (int)sizeof(*h)) 
		return 0;

	if(m->m_pkthdr.len < off+sizeof(struct udphdr)) {
		aspf_cnt_incomplete++;
		return 0;
	};

	m_copydata(m, off, sizeof(struct udphdr), (void*)&uh);

	ulen=ntohs(uh.uh_ulen);

	if(ulen < 8) { 
		/* by RFC 768: 
		Length  is the length  in octets  of this user datagram  including  
		this header  and the data. (This means the minimum value of the 
		length is eight.)
		*/
		aspf_cnt_incomplete++;
		goto drop;
	};

	if(ulen==8) { 
		/* This packet has zero payload */
		aspf_cnt_incomplete++;
		goto drop;
	};

	if(off+ulen > m->m_len) { 
		/* looks like we got partial packet */
		aspf_cnt_incomplete++;
		goto drop;
	};

	/*
	printf("ok, valid udp packet %i -> %i\n", htons(uh.uh_sport), 
		htons(uh.uh_dport));
	*/

	if(ntohs(uh.uh_dport) == aspf_dns_port) { 
		unsigned char buffer[ulen-8];
		m_copydata(m, off+8, ulen-8, buffer); 
		if(aspf_validate_dnsq(buffer, ulen-8)==0) { 
			aspf_invalid_dns_q++;
			goto drop;
		};
		aspf_valid_dns_q++;
		return 0;
	}; 
	return 0;
drop: 
	m_freem(m);
	*m0=NULL;
	return 1;
};

static int
aspf_hook(void)
{ 
	struct pfil_head* pfh_inet;
	pfh_inet = pfil_head_get(PFIL_TYPE_AF, AF_INET);
	if(pfh_inet == NULL) 
		return ENOENT;
	
	pfil_add_hook(aspf_check_in, NULL, PFIL_IN | PFIL_WAITOK, pfh_inet);

	return 0;
};

static int
aspf_unhook(void)
{ 
	struct pfil_head* pfh_inet;
	pfh_inet = pfil_head_get(PFIL_TYPE_AF, AF_INET);
	if(pfh_inet == NULL) 
		return ENOENT;
	
	pfil_remove_hook(aspf_check_in, NULL, PFIL_IN | PFIL_WAITOK, pfh_inet);

	return 0;
};

int
aspf_sysctl_hook(SYSCTL_HANDLER_ARGS)
{ 
	int enable=*(int*)arg1;
	int error;
	
	error=sysctl_handle_int(oidp, &enable, 0, req);

	if(error) 
		return error;

	enable=(enable) ? 1 : 0;

	if(enable==*(int*)arg1)
		return 0;

	if(arg1==&aspf_enabled) { 
		if(enable) 
			error=aspf_hook();
		else
			error=aspf_unhook();
	};

	if(error) 
		return error;

	*(int*)arg1=enable;
	
	return 0;
};

static int 
aspf_init(void)
{ 
	return 0;
};

int 
aspf_modevent(module_t md, int /*modeventtype_t*/ what, void* unused)
{ 
	int err=0;
	switch(what) { 
	case MOD_LOAD: 
		if((err=aspf_init())!=0) { 
			printf("aspf_init failed\n");
			break;
		};
		if((err=aspf_hook())!=0) { 
			printf("aspf_hook failed\n");
			break;
		};
		break;
	case MOD_UNLOAD: 
		if((err=aspf_unhook())!=0) { 
			printf("aspf_hook failed\n");
			break;
		};
		break;
	default : 
		return EOPNOTSUPP;
	};
	return err;
};

static moduledata_t aspfmod = { 
	"aspf",
	aspf_modevent,
	0
};

MODULE_VERSION(aspf, 0);
DECLARE_MODULE(aspf, aspfmod, SI_SUB_EXEC, SI_ORDER_ANY);

