/*-
 * Copyright (c) 2008, Alexandre Snarskii
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: aspf_validate.c,v 1.3 2008/10/27 13:35:47 snar Exp $
 */
#include <sys/types.h>
#include <sys/cdefs.h>
#include <sys/systm.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/kernel.h>
#include <sys/sysctl.h>

#include <netinet/in.h>

#include "aspf.h"

static unsigned long aspf_dnsq_invsize=0;
static unsigned long aspf_dnsq_responses=0;
static unsigned long aspf_dnsq_opcode=0;
static unsigned long aspf_dnsq_rcode=0;
static unsigned long aspf_dnsq_noquery=0;
static unsigned long aspf_dnsq_authsec=0;
static unsigned long aspf_dnsq_answersec=0;
static int aspf_dnsq_debug=0;

SYSCTL_DECL(_net_inet_aspf);
SYSCTL_NODE(_net_inet_aspf, OID_AUTO, dnsq, CTLFLAG_RW, 0, 
	"ASPF DNS Query filter statistics");
SYSCTL_ULONG(_net_inet_aspf_dnsq, OID_AUTO, badsize, CTLFLAG_RD, 
	&aspf_dnsq_invsize, 0, "DNS Query of invalid size");
SYSCTL_ULONG(_net_inet_aspf_dnsq, OID_AUTO, answer, CTLFLAG_RD, 
	&aspf_dnsq_responses, 0, "DNS Query with answer bit set");
SYSCTL_ULONG(_net_inet_aspf_dnsq, OID_AUTO, opcode, CTLFLAG_RD, 
	&aspf_dnsq_opcode, 0, "DNS Query with invalid opcode");
SYSCTL_ULONG(_net_inet_aspf_dnsq, OID_AUTO, rcode, CTLFLAG_RD, 
	&aspf_dnsq_rcode, 0, "DNS Query with non-zero Return Code");
SYSCTL_ULONG(_net_inet_aspf_dnsq, OID_AUTO, noquery, CTLFLAG_RD, 
	&aspf_dnsq_noquery, 0, "DNS Query with empty Query section");
SYSCTL_ULONG(_net_inet_aspf_dnsq, OID_AUTO, answersec, CTLFLAG_RD, 
	&aspf_dnsq_answersec, 0, "DNS Query with non-empty Answer section");
SYSCTL_ULONG(_net_inet_aspf_dnsq, OID_AUTO, authsec, CTLFLAG_RD, 
	&aspf_dnsq_authsec, 0, "DNS Query with non-empty Authority section");
SYSCTL_INT(_net_inet_aspf_dnsq, OID_AUTO, debug, CTLFLAG_RW, 
	&aspf_dnsq_debug, 0, "Enable DNS Query validation debug");

int
aspf_validate_dnsq(unsigned char* buffer, int len)
{ 
	unsigned char opcode;

	if(len<12 || len>512) { 
		/* RFC 1035, 4.2.1. UDP packets must have header (12 bytes) and
		restricted to 512 bytes (not counting IP and UDP headers) */
		aspf_dnsq_invsize++;
		if(aspf_dnsq_debug) 
			printf("aspf_validate_dnsq: invalid query size: %i\n", len);
		return 0;
	};

	if(buffer[2] & 0x80) { 
		/* it's a response, not a query */
		aspf_dnsq_responses++;
		return 0;
	};

	/* Allowed opcodes: Query, Status, Notify, Upadte. */
	/* Disallowed: all others (IQuery obsolete by RFC 3425) */
	opcode=(buffer[2]&0x78)>>3;
	if(opcode!=0 && opcode !=2 && opcode!=4 && opcode!=5) { 
		aspf_dnsq_opcode++;
		if(aspf_dnsq_debug) 
			printf("aspf_validate_dnsq: invalid opcode: %i\n", opcode);
		return 0;
	};

	/* Response code set in Query */
	opcode=(buffer[3]&0x0f);
	if(opcode) { 
		aspf_dnsq_rcode++;
		if(aspf_dnsq_debug) 
			printf("aspf_validate_dnsq: query rcode: %i\n", opcode);
		return 0;
	};

	/* No Query section in Query */
	if(!ntohs(*(uint16_t*)(buffer+4))) { 
		aspf_dnsq_noquery++;
		return 0;
	};

	/* Answer section present in Query */
	if(ntohs(*(uint16_t*)(buffer+6))) { 
		aspf_dnsq_answersec++;
		if(aspf_dnsq_debug) 
			printf("aspf_validate_dnsq: query has %i RR's in answer section\n", 
				ntohs(*(uint16_t*)(buffer+6)));
		return 0;
	};
	/* Authority section present in Query */
	if(ntohs(*(uint16_t*)(buffer+8))) { 
		aspf_dnsq_authsec++;
		if(aspf_dnsq_debug) 
			printf("aspf_validate_dnsq: query has %i records in authority"
				" section\n", ntohs(*(uint16_t*)(buffer+8)));
		return 0;
	};

	/* additional records is OK, used for DNSSEC 
	if(ntohs(*(uint16_t*)(buffer+10))) { 
		printf("validate_dns_q: %i additional RR's in query\n", 
			ntohs(*(uint16_t*)(buffer+10)));
		return 0;
	};
	*/

	return 1;
};
	
